use crate::error::*;
use crate::usimp;
use crate::usimp::*;
use futures::stream::{SplitSink, SplitStream};
use futures_util::SinkExt;
use futures_util::StreamExt;
use hyper::{header, Body, Request, StatusCode};
use hyper_tungstenite::hyper::upgrade::Upgraded;
use hyper_tungstenite::tungstenite::{handshake, Message};
use hyper_tungstenite::{tungstenite::protocol::Role, WebSocketStream};
use serde_json::{Map, Value};
use tokio::sync::mpsc;

async fn sender(
    mut sink: SplitSink<WebSocketStream<Upgraded>, Message>,
    mut rx: mpsc::Receiver<OutputEnvelope>,
) {
    while let Some(msg) = rx.recv().await {
        let mut envelope = Value::Object(Map::new());
        envelope["data"] = msg.data;
        envelope["request_nr"] = match msg.request_nr {
            Some(nr) => Value::from(nr),
            None => Value::Null,
        };
        match msg.error {
            Some(error) => {
                envelope["status"] = Value::from("error");
                envelope["error"] = Value::from(error);
            }
            None => {
                envelope["status"] = Value::from("success");
            }
        }
        if let Err(error) = sink.send(Message::Text(envelope.to_string())).await {
            eprintln!("{:?}", error);
            break;
        }
    }
}

async fn receiver(
    mut stream: SplitStream<WebSocketStream<Upgraded>>,
    tx: mpsc::Sender<OutputEnvelope>,
) {
    while let Some(res) = stream.next().await {
        match res {
            Ok(msg) => {
                let input: InputEnvelope = serde_json::from_slice(&msg.into_data()[..]).unwrap();
                let output = match usimp::endpoint(&input, Some(tx.clone())).await {
                    Ok(output) => output,
                    Err(error) => input.error(error),
                };
                let _res = tx.send(output).await;
            }
            Err(error) => println!("{:?}", error),
        }
    }
}

pub async fn handler(
    req: Request<Body>,
    res: hyper::http::response::Builder,
) -> (hyper::http::response::Builder, Option<OutputEnvelope>) {
    match req.headers().get(header::UPGRADE) {
        Some(val) if val == header::HeaderValue::from_str("websocket").unwrap() => {}
        _ => {
            return (
                res,
                Some(OutputEnvelope::from(Error::new(
                    ErrorKind::WebSocketError,
                    ErrorClass::ClientProtocolError,
                    None,
                ))),
            )
        }
    }

    let key = match req.headers().get(header::SEC_WEBSOCKET_KEY) {
        Some(key) => key,
        None => {
            return (
                res,
                Some(OutputEnvelope::from(Error::new(
                    ErrorKind::WebSocketError,
                    ErrorClass::ClientProtocolError,
                    None,
                ))),
            )
        }
    };
    let key = handshake::derive_accept_key(key.as_bytes());

    match req.headers().get(header::SEC_WEBSOCKET_PROTOCOL) {
        Some(val) if val == header::HeaderValue::from_str("usimp").unwrap() => {}
        _ => {
            return (
                res,
                Some(OutputEnvelope::from(Error::new(
                    ErrorKind::UsimpError,
                    ErrorClass::ClientProtocolError,
                    None,
                ))),
            )
        }
    }

    tokio::spawn(async move {
        match hyper::upgrade::on(req).await {
            Ok(upgraded) => {
                let ws_stream =
                    WebSocketStream::from_raw_socket(upgraded, Role::Server, None).await;
                let (tx, rx) = mpsc::channel::<OutputEnvelope>(64);
                let (sink, stream) = ws_stream.split();
                tokio::spawn(async move { sender(sink, rx).await });
                receiver(stream, tx).await
            }
            Err(error) => eprintln!("Unable to upgrade: {}", error),
        }
    });

    (
        res.status(StatusCode::SWITCHING_PROTOCOLS)
            .header(
                header::CONNECTION,
                header::HeaderValue::from_str("Upgrade").unwrap(),
            )
            .header(
                header::UPGRADE,
                header::HeaderValue::from_str("websocket").unwrap(),
            )
            .header(
                header::SEC_WEBSOCKET_ACCEPT,
                header::HeaderValue::from_str(key.as_str()).unwrap(),
            )
            .header(
                header::SEC_WEBSOCKET_PROTOCOL,
                header::HeaderValue::from_str("usimp").unwrap(),
            ),
        None,
    )
}
